module AhornFavoriteSaver

using ..Ahorn, Maple, Gtk, Gtk.ShortNames, Ahorn.Form

displayName = "Favorites"
group = "Favorites"

favoritesLayerList = [
    "favorites_placements_entities",
    "favorites_placements_triggers",
    "favorites_placements_fgDecals",
    "favorites_placements_bgDecals",
    "favorites_brushes_fgTiles",
    "favorites_brushes_bgTiles",
    "favorites_objtiles_objtiles",
]

lastFavoritesWindow = nothing
lastFavoritesWindowDestroyed = false

function toolSelected(subTools::Ahorn.ListContainer, layers::Ahorn.ListContainer, materials::Ahorn.ListContainer)
    Ahorn.updateLayerList!(String[])

    updateMaterialsList!()
end

function materialSelected(list::Ahorn.ListContainer, selected::String)
    saveFavorites()
    
    if selected == "(New)"
        currentSet = getCurrentSet()
        Ahorn.selectRow!(list, row -> row[1] == currentSet)
        displayFavoritesCreator()
    else
        Ahorn.persistence["vitellary_current_favorites_set"] = selected
        loadFavorites()
    end
end

function materialDoubleClicked(material::String)
    if material != "(Default)" && material != "(New)"
        displayFavoritesEditor(material)
    end
end

function updateMaterialsList!()
    selectables = collect(keys(getFavoritesList()))
    sort!(selectables)
    selectables = [selectables; "(New)"]
    
    currentSet = getCurrentSet()
    Ahorn.setMaterialList!(selectables, row -> row[1] == currentSet)
end

function getFavoritesList()
    favorites = get(Ahorn.persistence, "vitellary_favorites_sets", Dict{String, Any}(
        "(Default)" => Dict{String, Any}()
    ))
    Ahorn.persistence["vitellary_favorites_sets"] = favorites

    return favorites
end

function getCurrentSet()
    return get(Ahorn.persistence, "vitellary_current_favorites_set", "(Default)")
end

function saveFavorites()
    favorites = getFavoritesList()
    currentSet = getCurrentSet()
    favorites[currentSet] = Dict{String, Any}()
    for key in favoritesLayerList
        favorites[currentSet][key] = deepcopy(Ahorn.getFavorites(Ahorn.persistence, key))
    end
end

function loadFavorites()
    favorites = getFavoritesList()
    currentSet = getCurrentSet()
    newSet = get(favorites, currentSet, Dict{String, Any}())
    for key in favoritesLayerList
        Ahorn.persistence[key] = deepcopy(get(newSet, key, Array{String, 1}()))
    end
end

function createFavoritesFormWindow(title::String, sections::Union{Array{Form.Section, 1}, Form.Section}; columns::Integer=4, separateGroups::Bool=true,
    gridIfSingleSection::Bool=true, buttonText::String="Add", callback::Function=(name) -> println(name), removeCallback::Union{Function, Nothing}=nothing,
    parent::Gtk.GtkWindow=Ahorn.window, icon::Pixbuf=Ahorn.windowIcon, canResize::Bool=false)

    sections = isa(sections, Form.Section) ? Form.Section[sections] : sections
    finishButton = Form.Button(buttonText)
    removeButton = Form.Button("Remove")

    content = Form.generateSectionsNotebook(sections, columns=columns, separateGroups=separateGroups, gridIfSingleSection=gridIfSingleSection)

    window = Form.Window(title, -1, -1, canResize, icon=icon) |> (Form.Frame() |> (box = Form.Box(:v)))
    
    push!(box, content)
    push!(box, finishButton)
    if removeCallback != nothing
        push!(box, removeButton)
    end

    if removeCallback != nothing
        @Ahorn.guarded signal_connect(removeButton, "clicked") do args...
            remove = Ahorn.topMostAskDialog("Really remove this set?", lastFavoritesWindow)

            if remove
                removeCallback()
            end
        end
    end

    @Ahorn.guarded signal_connect(finishButton, "clicked") do args...
        data, incorrectOptions = Form.getSectionsData(sections)

        if data !== nothing
            callback(get(data, "name", ""))

        else
            Ahorn.topMostInfoDialog(Form.getIncorrectOptionsMessage(incorrectOptions), window)
        end
    end

    return window
end

function spawnFavoritesWindow(name::String, buttonText::String, title::String, callback::Function, removeCallback::Union{Function, Nothing}=nothing)
    #keepPreviousPosition = get(Ahorn.config, "properties_keep_previous_position", true)
    keepPreviousPosition = true

    winX, winY = 0, 0
    winScreen = nothing

    options = Form.Option[Form.suggestOption("Name", name, dataName="name", editable=true)]

    section = Form.Section("properties", options)
    favoritesWindow = createFavoritesFormWindow(title, section, buttonText=buttonText, callback=callback, removeCallback=removeCallback)
    @Ahorn.guarded signal_connect(favoritesWindow, :destroy) do widget
        global lastFavoritesWindowDestroyed = true
    end

    visible(favoritesWindow, false)

    if keepPreviousPosition && isa(lastFavoritesWindow, Gtk.GtkWindowLeaf) && !lastFavoritesWindowDestroyed
        winX, winY = GAccessor.position(lastFavoritesWindow)
        winScreen = GAccessor.screen(lastFavoritesWindow)

        GAccessor.position(favoritesWindow, winX, winY)
        GAccessor.screen(favoritesWindow, winScreen)
    end

    GAccessor.transient_for(favoritesWindow, Ahorn.window)

    showall(favoritesWindow)
    visible(favoritesWindow, true)

    if isa(lastFavoritesWindow, Gtk.GtkWindowLeaf)
        Gtk.destroy(lastFavoritesWindow)
    end

    global lastFavoritesWindow = favoritesWindow
    global lastFavoritesWindowDestroyed = false
end

function displayFavoritesEditor(set::String)
    removeCallback = function()
        saveFavorites()

        favorites = getFavoritesList()
        delete!(favorites, set)

        Ahorn.persistence["vitellary_current_favorites_set"] = "(Default)"

        newSet = get(favorites, "(Default)", Dict{String, Any}())
        loadFavorites()

        updateMaterialsList!()
        Gtk.destroy(lastFavoritesWindow)
    end

    callback = function(name::String)
        if name == "" || name == "(New)" || name == set
            return
        end

        update = true
        favorites = getFavoritesList()

        if haskey(favorites, name)
            update = Ahorn.topMostAskDialog("There is already a set with this name\nDo you want to overwrite it?", lastFavoritesWindow)
        end

        if update
            saveFavorites()

            favorites[name] = get(favorites, set, Dict{String, Any}())
            delete!(favorites, set)

            Ahorn.persistence["vitellary_current_favorites_set"] = name
            loadFavorites()

            updateMaterialsList!()
            Gtk.destroy(lastFavoritesWindow)
        end
    end

    spawnFavoritesWindow(set, "Rename", "Editing favorites set - $(set)", callback, removeCallback)
end

function displayFavoritesCreator()
    callback = function(name::String)
        if name == "" || name == "(Default)" || name == "(New)"
            return
        end

        update = true
        favorites = getFavoritesList()

        if haskey(favorites, name)
            update = Ahorn.topMostAskDialog("There is already a set with this name\nDo you want to overwrite it?", lastFavoritesWindow)
        end

        if update
            saveFavorites()
            favorites[name] = deepcopy(get(favorites, "(Default)", Dict{String, Any}()))

            Ahorn.persistence["vitellary_current_favorites_set"] = name
            loadFavorites()

            updateMaterialsList!()
            Gtk.destroy(lastFavoritesWindow)
        end
    end

    spawnFavoritesWindow("", "Create", "Creating favorites set", callback)
end

end